function [ struct_data, struct_header ] = NE_LoadNedf( nedfPath )
%----------------------------------------------------------
% EEGStarlab Toolbox
% (C) Neuroelectrics Barcelona S.L.
% anton.albajes-eizagirre@neuroelectrics.com
%----------------------------------------------------------
% Loads a NEDF file with EEG and/or Stim data into two structs with signals and header
%
% This function loads a Neuroelectric's NEDF file and returns two structs, one with 
% the data signals, and the other with the NEDF file. Works with both Enobio (EEG)
% and Starstim (Stim) devices files. 
%
%----------------------------------------------------------
%Inputs
% nedfPath: string containing the absolute path to the nedf file to load.
%----------------------------------------------------------
%Outputs
% struct_data: struct with data from the NEDF eeg streams. Struct contains:
%   - EEG: SDC with EEG stream
%   - Stim: SDC with Stim stream
%   - External: SDC with external stream
%   - Accelerometer: SDC with accelerometer stream
% struct_header: struct with the content of the xml header
%----------------------------------------------------------
% Dependency: XML4MAT http://www.mathworks.com/matlabcentral/fileexchange/6268-xml4mat-v2-0
%----------------------------------------------------------
% Version   Date        Author  Changes 
% v1        2014/10/10  AAE     Initial version 
% v1.1      2015/01/25  AAE     Fixed sample scale bug and performance
%                               optimisation
% v1.2      2015/01/30  AAE     Changed function/file name to NE_LoadNedf
%----------------------------------------------------------
% EX.
%
% [data,header] = NeuroLoadNedf('/home/neuroelectrics/recording.nedf')
%----

% Here goes the code of the function
fileId = fopen(nedfPath);
A = fread(fileId); % read file as a column of bytes
foo = find(A==0); % PRESUMABLY, bytes = 0 are binary values (end of XML)
xml = A(1:foo(1));
binStart = 5120; % binary data will always start at byte 5210
binData = A(binStart+1:end);
binHeader = 1;
addpath('xml4mat');
[struct_header, var_name] = xml2mat(char(xml)');
struct_header.RecordName = var_name;
nChanEeg = str2num(struct_header.NumberOfEEGChannels);
nChan = str2num(struct_header.TotalNumberOfChannels);
% check if there's an Accelerometer data stream in the NEDF file
if isfield(struct_header,'NumberOfChannelsOfAccelerometer')
    nChanAcc = str2num(struct_header.NumberOfChannelsOfAccelerometer);
else
    nChanAcc=0;
end
nChanExt = 0;
% check if there's an external channel data stream in the NEDF file
if(struct_header.AdditionalChannelStatus~='OFF')
    nChanExt = 1;
end
deviceClass = struct_header.DeviceClass; 
counterEeg = 1;
counterStim = 1;
counterExt = 1;
counterTrig = 1;
counterAcc = 5;
lastSampleAcc=zeros(nChanAcc,1);
nSamplesStim = 0;
% if the device is a StarStim, with Stim data stream, we'll get two samples
% of stim for each one of EEG
if strcmp(deviceClass,'StarStim')
    nSamplesStim = 2;
end
nRecordsEeg = str2num(struct_header.NumberOfRecordsOfEEG);
% check if the header includes a stim data stream. If so, get how many
% records the stim data stream has
% if stim stream is not included in the header, set number of samples of
% stim to get by iteration to 0.
if isfield(struct_header,'NumberOfRecordsOfStimulation')
   nRecordsStim = str2num(struct_header.NumberOfRecordsOfStimulation);
else
   nRecordsStim = 2; % we set number a fake number for stim records, so it won't interfere with the EEG loading with an early stop. 
   nSamplesStim = 0;
end
struct_data.Triggers = [];
% flag that will be used to check when one of the streams (EEG or Stim) has
% come to an end
moreData = 1;
struct_data.Accelerometer = zeros(nRecordsEeg-1,nChan);
struct_data.EEG = zeros(nRecordsEeg-1,nChan);
struct_data.Stim = zeros((nRecordsEeg-1)*nSamplesStim,nChan);
struct_data.Triggers = zeros(nRecordsEeg-1,1);
struct_data.External = zeros(nRecordsEeg-1,nChanExt);
while(moreData)
    % there will be one sample in the accelerometer stream per 5 samples in
    % the EEG stream
   if counterAcc == 5
       counterAcc = 1;
    for a= 1:nChanAcc
       [lastSampleAcc(a), binHeader] = getSampleAcc(binData,binHeader);
    end
   else
       counterAcc = counterAcc + 1;
   end
   for a=1:nChanAcc
       struct_data.Accelerometer(counterEeg,a) = lastSampleAcc(a);
   end

   % read EEG sample
   for e=1:nChan
       [sample,binHeader] = getSampleEEG(binData,binHeader);
       struct_data.EEG(counterEeg,e) = sample;
   end
   counterEeg = counterEeg+1;
   % if we already read all the available EEG records, stop the reading
   % loop
   if counterEeg >= nRecordsEeg
       moreData = 0;
   end
   
   % there will be nSamplesStim in the stim stream per one sample in the
   % EEG stream (currently 2 stim / 1 eeg)
   for ns = 1:nSamplesStim
       for s=1:nChan
           [sample,binHeader] = getSampleStim(binData,binHeader);
           struct_data.Stim(counterStim,s) = sample;
       end
       counterStim = counterStim+1;
   end
   % if we already read all the available Stim records, stop the
   % reading loop
   if counterStim >= nRecordsStim
       moreData = 0;
   end
   % if there is data in the external channel stream, read nChan samples
   for x=1:nChanExt
       [sample,binHeader] = getSampleEEG(binData,binHeader);
       struct_data.External(counterExt,x) = sample;
   end
   counterExt = counterExt +1;
   % load one-byte integer from the trigger stream
   struct_data.Triggers(counterTrig,1) = int8(binData(binHeader));
   counterTrig = counterTrig +1;
   binHeader = binHeader+1;
    
end

end

% function that will read one sample of one channel of the accelerometer
% stream.. data will be in three 
function [sample, header] = getSampleAcc(data, header)
   byte1 = int32(uint8(data(header)));
   header=header+1;
   byte2 = int32(uint8(data(header)));
   header = header+1;
   byte3 = int32(uint8(data(header)));
   header = header+1;
%   disp(['Byte3 ' num2str((byte3))]);  
   sample = byte1*65536+byte2*256+byte3;
   if(byte1>=128)
      sample = (sample - 16777216);
   end
end

% function that will read one sample of one channel of the eeg stream
% data will be in three bytes 2-complemented and then scaled
function [sample, header] = getSampleEEG(data, header)
   byte1 = double(uint8(data(header)));
   header=header+1;
   byte2 = double(uint8(data(header)));
   header = header+1;
   byte3 = double(uint8(data(header)));
   header = header+1;
   
   sample = byte1*65536+byte2*256+byte3;
   if(byte1>=128)
      sample = (sample - 16777216); 
   end
   sample = sample*(2.4*1000000000.0/6.0/8388607.0);
end

% function that will read one sample of one channel of the stimulation stream
% data will be in three bytes 2-complemented. According to sampled NEDF
% files, data will not be scaled.
function [sample, header] = getSampleStim(data, header)
   byte1 = double(uint8(data(header)));
   header=header+1;
   byte2 = double(uint8(data(header)));
   header = header+1;
   byte3 = double(uint8(data(header)));
   header = header+1;
   
   sample = byte1*65536+byte2*256+byte3;
   if(byte1>=128)
      sample = (sample - 16777216); 
   end
   sample = sample;
end
